/*
** pc-video.c -- IBM PC image display routines
**
** Copyright 1992, 1993, 1994 by
** Michael J. Gourlay
**
** NO WARRANTEES, EXPRESS OR IMPLIED.
*/

#include "showtime.h"
#include "video.h"
#include "cmap.h"

/* ---------------------------------------------------------------------- */
#if defined(MSC)
#define MOVIE_PUT_IMAGE(x, y,img) _putimage(x, y, img, _GPSET)
#elif defined(TC)
#define MOVIE_PUT_IMAGE(x, y, img) putimage(x, y, img, COPY_PUT)
#endif /* architecture */

/* ====================================================================== */
/* Global variables */
/* gm_xoff, gm_yoff are the image offset */
int   gm_xoff=0, gm_yoff=0;

/* gm_xscale, gm_yscale are the image scale parameters */
float gm_xscale=1.0, gm_yscale=1.0;

/* actual window size */
int w_width=0, w_height=0;

/* ---------------------------------------------------------------------- */
/* tmpgroup is temporary image space that would be malloc'd/free'd often */
/* It's static to save time */
static unsigned char *tmpgroup=NULL; /* GLOBAL to this module */






















/* ====================================================================== */

/* initialize_display : initialize graphics mode with min size widthXheight
** Set the graphics mode
** a nonresizable window of at least width and height is made if possible
**
** Also allocate memory for xscale
** GLOBAL to this module, non dev specific: tmpgroup
*/
void
initialize_display(int width, int height)
{
#if defined(MSC)
  struct _videoconfig vcfg;
  int minmode;

  if(width <= 320 && height <= 200) minmode=_MRES256COLOR;
  else if(width <= 640 && height <= 400) minmode=_ORES256COLOR;
  else if(width <= 640 && height <= 480) minmode=_VRES256COLOR;
  else if(width <= 800 && height <= 600) minmode=_SRES256COLOR;
  else if(width <= 1024 && height <= 768) minmode=_XRES256COLOR;
  else minmode=_ZRES256COLOR;

  switch(minmode) {
    case _MRES256COLOR:
      if(! _setvideomode(_MRES256COLOR)) {
        logmesg("\ninit_disp: could not open MRES256COLOR mode\n");
        bye_bye(1);
      } else {
        w_width = 320;
        w_height = 200;
      }
      break;

    case _ORES256COLOR:
      if(! _setvideomode(_ORES256COLOR)) {
        logmesg("\ninit_disp: could not open ORES256COLOR mode\n");
        bye_bye(1);
      } else {
        w_width = 640;
        w_height = 400;
      }
      break;

    case _VRES256COLOR:
      if(! _setvideomode(_VRES256COLOR)) {
        logmesg("\ninit_disp: could not open VRES256COLOR mode\n");
        bye_bye(1);
      } else {
        w_width = 640;
        w_height = 480;
      }
      break;

    case _SRES256COLOR:
      if(! _setvideomode(_SRES256COLOR)) {
        logmesg("\ninit_disp: could not open SRES256COLOR mode\n");
        bye_bye(1);
      } else {
        w_width = 800;
        w_height = 600;
      }
      break;

    case _XRES256COLOR:
      if(! _setvideomode(_XRES256COLOR)) {
        logmesg("\ninit_disp: could not open XRES256COLOR mode\n");
        bye_bye(1);
      } else {
        w_width = 1024;
        w_height = 768;
      }
      break;

    case _ZRES256COLOR:
      if(! _setvideomode(_ZRES256COLOR)) {
        logmesg("\ninit_disp: could not open ZRES256COLOR mode\n");
        bye_bye(1);
      } else {
        w_width = 1280;
        w_height = 1024;
      }
      break;
  }

  _getvideoconfig(&vcfg);

  /*logmesg("video pages=%i\n", vcfg.numvideopages);/**/

#elif defined(TC)

  int gdriver = IBM8514;
  int errcode;

  initgraph(&gdriver, &GMODE_V, "");
  errcode = graphresult();

  if(errcode != grOk) {
    logmesg("\ninit_disp: graphics error %s\n", grapherrormsg(errcode));
    bye_bye(1);
  } else {
    w_width = 1280;
    w_height = 1024;
  }
#endif /* architecture */

  /* tmpgroup is GLOBAL to module */
  /* Allocate some memory for the image manipulation buffer */
  if((tmpgroup=malloc((size_t)128 + (size_t)w_width))==NULL) {
    logmesg("initialize_display: could not allocate xscale buffer\n");
  }
}

/* ---------------------------------------------------------------------- */
/* Clear the window */
void
clear_display(int bgcolor)
{
#if defined (MSC)
  _setcolor(bgcolor);
  _rectangle(_GFILLINTERIOR, 0, 0, w_width-1, w_height-1);
#endif

}










/* ---------------------------------------------------------------------- */

/* reset_display : put the graphics mode into a normal text mode
*/
void
reset_display(void)
{
#if defined(MSC)
  _setvideomode(_DEFAULTMODE);
#elif defined(TC)
  closegraph();
#endif /* architecture */
  if(tmpgroup!=NULL) free(tmpgroup); /* GLOBAL to module */
}

/* ====================================================================== */



/* show_image : scale and display an image loaded into memory
*/
/* also assumes that header_size will be the same as old one
** see the load routine for comments on header_size -- MJG 13Jan94
*/
void
show_image(screendata_t *sdP)
{
  register int   src_row;
  register int   prev_src_row;
  register int   xsize; /* screen (destination) image width */
  register int   ysize; /* screen (destination) image height */
  register int   xstep;
  register int   ystep;
  register int   hi;
  register int   vi;
  register int   ngi;
  unsigned char  *newgroup=NULL;
  unsigned char  *sd_row;

  if(sdP==NULL) return;

  if(gm_xscale!=1) {
    newgroup=tmpgroup; /* tmpgroup is GLOBAL to module */
    /* copy old header from row 0 (assumes all headers are identical) */
    memcpy(/*to*/newgroup, sdP->data[0], (size_t)sdP->header_size);
  }

  if(gm_xscale>0) {
    xsize = sdP->num_cols * gm_xscale;
    /* modify "width" field of new row header */
    ((UWORD *)newgroup)[0] = (UWORD)(xsize * COL_FACTOR);
  } else if(gm_xscale<0) {
    xstep = -gm_xscale;
    xsize = sdP->num_cols / xstep;
    /* modify "width" field of new row header */
    ((UWORD *)newgroup)[0] = (UWORD)(xsize * COL_FACTOR);
  } else {
    xsize = sdP->num_cols;
  }

  if(gm_yscale>0) {
    ysize = sdP->num_rows * gm_yscale;
  } else if(gm_yscale<0) {
    ystep = -gm_yscale;
    ysize = sdP->num_rows / ystep;
  } else {
    ysize = sdP->num_rows;
  }

  /* Clip the image if it overwrites the window */
  if(xsize + gm_xoff > w_width) xsize = w_width - gm_xoff;
  if(ysize + gm_yoff > w_height) ysize = w_height - gm_yoff;
  if(xsize<1 || ysize<1) return;

  /* Display each row of the image */
  prev_src_row = -1;
  for(vi=0; vi<ysize; vi++) {

    if(gm_yscale > 0) {
      src_row = (int)(vi/gm_yscale);
    } else {
      src_row = vi*ystep;
    }
    sd_row = sdP->data[src_row];

    if(src_row != prev_src_row) {
      /* Scale horizontally */
      if(gm_xscale!=1 && gm_xscale>0) {
        for(hi=0; hi<xsize; hi++) {
          newgroup[hi+sdP->header_size] = sd_row[sdP->header_size + (int)(hi/gm_xscale)];
        }
      } else if(gm_xscale!=1 && gm_xscale<0) {
        /* shrink integrally */
        for(hi=0; hi<xsize; hi++) {
          newgroup[hi+sdP->header_size] = sd_row[sdP->header_size + hi*xstep];
        }
      } else {
        newgroup=sd_row;
      }

    }
    /* Put the image up on the graphics screen */
    MOVIE_PUT_IMAGE(gm_xoff, gm_yoff+vi, newgroup);

    prev_src_row=src_row;
  } /* for each row of the screen (destination) image */
}
