/*
** pic.c -- PIC and CLP file handling routines for ShowTime
**
** Copyright (C) 1994 by Michael J. Gourlay
**
** NO WARRANTEES, EXPRESS OR IMPLIED.
**
*/

#include <stdio.h>
#include <stdlib.h>

#include "showtime.h"
#include "pic.h"
#include "cmap.h"
#include "file.h"

/* RGB color index offsets into a colormap table */
#define R 0
#define G 1
#define B 2

/* ====================================================================== */

/* load_pic_header : read a PIC image header and return pertinent info
*/
int
load_pic_header(image_hdr_t *ihdrP, screendata_t *sdP, FILE *fio)
{
  pic_hdr_t *picP;

  /* File is probably a PIC or CLP file */
  if((picP = ihdrP->hdr = (pic_hdr_t *)malloc(sizeof(pic_hdr_t)))==NULL) {
    logmesg("load_pic_header: out of mem allocating picP\n");
    exit(1);
  }

  /* Read a PIC header */
  NONEGB(ihdrP->num_cols, "rih: PIC num_cols"); /* width */
  NONEGB(ihdrP->num_rows, "rih: PIC num_rows"); /* height */

  NONEGB(picP->x_off,"rih: PIC xoff"); /* horizontal offset */
  NONEGB(picP->y_off,"rih: PIC yoff"); /* vertical offset */
  sdP->xoffset = picP->x_off;
  sdP->yoffset = picP->y_off;

  /* bits per pixel */
  GET_BYTE(fio, picP->bpp);
  picP->planes = ((picP->bpp & 0xf0) >> 4) + 1;

  /* read marker */
  GET_BYTE(fio, picP->marker);  /* should be 255 */

  /* read video type */
  GET_BYTE(fio, picP->vtype);

  /* read palette info */
  picP->etype=get_le_word(fio); /* palette type */
  picP->elen = get_le_word(fio); /* number of palette bytes */

  /* Now, we read the palette if there is one */
  if(picP->elen) {
    load_pic_palette(fio, picP->etype, picP->elen);
  }

  ihdrP->load_img = load_pic_image;

  return(0);
}

/* ---------------------------------------------------------------------- */

int
load_pic_palette(FILE *fio, int edesc, int elen)
{
  int bcol, cgapal;
  int i;
  unsigned char rgb_cmap[256][3];
  unsigned char *buf;

  switch (edesc) {
    case 0:
      break;
    case 1: /* palette number + Border color (2 bytes) */

      if(elen!=2) {
        logmesg("load_pic_pal: unexpected cmap size in PIC CGA palette\n");
      } else {

        GET_BYTE(fio, cgapal);
        GET_BYTE(fio, bcol);

        /* border color is a CGA color number */
        rgb_cmap[0][R] = cga_color[bcol][R];
        rgb_cmap[0][G] = cga_color[bcol][G];
        rgb_cmap[0][B] = cga_color[bcol][B];

        /* Choose among the lame 4-color CGA palettes */
        for (i=0;i<3;i++) {
          rgb_cmap[i][R] = cga_color[cga_palette[cgapal][i]][R];
          rgb_cmap[i][G] = cga_color[cga_palette[cgapal][i]][G];
          rgb_cmap[i][B] = cga_color[cga_palette[cgapal][i]][B];
        }

        /* install palette */
        install_cmap(4, rgb_cmap, 255);
      }
      break;


    case 2: /* PCjr colors.  16 1 byte color numbers */
      if(elen!=16) {
        logmesg("load_pic_pal: unexpected cmap size in PIC PCjr palette\n");
      }
      if((buf=malloc(elen))==NULL) {
        logmesg("load_pic_pal: could not malloc %i bytes for palette\n", elen);
        bye_bye(1);
      }
      if(get_block(fio, buf, elen)) {
        logmesg("load_pic_pal: read EOF reading PIC PCjr palette");
        return(EOF);
      }
      for (i=0;i<elen;i++) {
        rgb_cmap[i][R] = cga_color[buf[i]][R];
        rgb_cmap[i][G] = cga_color[buf[i]][G];
        rgb_cmap[i][B] = cga_color[buf[i]][B];
      }
      free(buf);

      /* install palette */
      install_cmap(elen, rgb_cmap, 255);
      break;


    case 3: /* EGA 16/64.  16 numbers in range 0-63. */
      if(elen!=16) {
        logmesg("load_pic_pal: unexpected cmap size in PIC EGA palette\n");
      }
      if((buf=malloc(elen))==NULL) {
        logmesg("load_pic_pal: could not malloc %i bytes for palette\n", elen);
        bye_bye(1);
      }
      if (get_block(fio, buf, elen)) {
        logmesg("load_pic_pal: read EOF reading PIC EGA palette\n");
        return(EOF);
      }
      for (i=0;i<elen;i++) {
#ifdef EGA_ENABLED
        rgb_cmap[i][R] = ega_color[buf[i]][R];
        rgb_cmap[i][G] = ega_color[buf[i]][G];
        rgb_cmap[i][B] = ega_color[buf[i]][B];
#else
        logmesg("load_pic_pal: EGA colors not supported;  faking it\n");
        rgb_cmap[i][R] = cga_color[buf[i]][R];
        rgb_cmap[i][G] = cga_color[buf[i]][G];
        rgb_cmap[i][B] = cga_color[buf[i]][B];
#endif
      }
      free(buf);

      /* install palette */
      install_cmap(elen, rgb_cmap, 255);
      break;


    case 5: /* Super EGA color map */
    case 4: /* VGA color map */


      /* Read cmap */
      if(elen != 256*3) {
        logmesg("load_pic_pal: unexpected cmap size in PIC VGA palette\n");
      }
      if(get_block(fio, (unsigned char *)rgb_cmap, elen)) {
        logmesg("load_pic_pal: unexpected EOF reading PIC VGA palette");
        return(EOF);
      }

      /* install palette */
      install_cmap(elen/3, rgb_cmap, 63);
      break;


    default:
      logmesg("load_pic_pal: unrecognized extended PIC info type %d.\n", edesc);
      logmesg("load_pic_pal: Skipping %d bytes...\n", elen);
      if (!elen)
        break;
      if((buf=malloc(elen))==NULL) {
        logmesg("load_pic_pal: could not malloc %i bytes for palette\n", elen);
        bye_bye(1);
      }
      if (get_block(fio, buf, elen)) {
        logmesg("load_pic_pal: unexpected EOF reading PIC palette");
        return(1);
      }
      free(buf);
  }

  return(0);
}

/* ---------------------------------------------------------------------- */

/* load_pic_image: read a PIC or CLP image into sdP from fio
*/
int
load_pic_image(image_hdr_t *ihdrP, screendata_t *sdP, FILE *fio)
{
  unsigned char esc_byte;
  unsigned char pic_byte;
  unsigned char rle_count_byte;
  register int  row;
  register int  col;
  register int  i;
  register long ncols, hsize; /* local versions for speed */
  register long rle_count;
  long          block_size;
  long          pic_blocks;
  long          rle_bytes;
  long          image_size;
  long          image_byte=0;

  /* sdP should have been set in image header load routine  */
  if(sdP==NULL) {
    logmesg("\nload_pic_image: screen data is null\n");
    return(1);
  }

  ncols=sdP->num_cols; /* for speed in loop */
  hsize=sdP->header_size; /* for speed in loop */

  image_size = sdP->num_rows * sdP->num_cols;

  /* --- Load the image --- */
  /* Find the number of compressed blocks */
  NONEGB(pic_blocks, "PIC blocks");

  while(pic_blocks>0) {
    /* Read a block head */
    NONEGB(rle_bytes, "PIC RLE bytes");   /* compressed block size */
    NONEGB(block_size, "PIC block size"); /* decompressed size */
    GET_BYTE(fio, esc_byte);

    /* rle_bytes includes the size of the header, so subtract that */
    rle_bytes -= 5;

    /* Read a block */
    while(rle_bytes>0) {
      /* read PIC byte */
      GET_BYTE(fio, pic_byte);

      rle_bytes--;

      /* See if this byte is the beginning of an RLE run */
      if(pic_byte == esc_byte) {
        /* This is an RLE run, so unpack it */
        /* Get the repeate count byte */
        GET_BYTE(fio, rle_count_byte);

        rle_bytes--;

        /* If the RLE count is zero, the actual, 2-byte count follows */
        if(rle_count_byte == 0) {
          /* read the 2-byte RLE count */
          NONEGB(rle_count, "PIC RLE count");

          rle_bytes -= 2;
        } else {
          rle_count = rle_count_byte;
        }

        /* read the literal PIC byte */
        GET_BYTE(fio, pic_byte);

        rle_bytes--;

        /* Store the PIC byte in the image data */
        for(i=0; i<rle_count; i++) {
          col = image_byte % ncols + hsize;
          row = sdP->num_rows - 1 - image_byte / ncols;
          sdP->data[row][col] = pic_byte;
          image_byte++;
        }

      } else {
        /* This is not an RLE run;  Use this byte literally  */
        col = image_byte % ncols + hsize;
        row = sdP->num_rows - 1 - image_byte / ncols;
        sdP->data[row][col] = pic_byte;
        image_byte++;
      }

#ifdef TEST_SIZE
      /* Check for consistency of size */
      if(image_byte > image_size) {
        logmesg("\nload_pic_image: decompression is too large\n");
        logmesg("\nload_pic_image: %li > %li\n", image_byte, image_size);
        return(2);
      }
#endif /* TEST_SIZE */

    } /* read a block */

    pic_blocks--;

  } /* read all blocks */

  return(0);
}
