/*
** x-video.c -- X Window System image display routines
**
** Copyright 1992, 1993, 1994 by
** Michael J. Gourlay
**
** NO WARRANTEES, EXPRESS OR IMPLIED.
*/

#include "showtime.h"
#include "video.h"
#include "cmap.h"

/* ---------------------------------------------------------------------- */

#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>


/* ====================================================================== */
/* Global variables */
/* gm_xoff, gm_yoff are the image offset */
int   gm_xoff=0, gm_yoff=0;

/* gm_xscale, gm_yscale are the image scale parameters */
float gm_xscale=1.0, gm_yscale=1.0;

/* actual window size */
int w_width=0, w_height=0;

/* ---------------------------------------------------------------------- */
/* tmpgroup is temporary image space that would be malloc'd/free'd often */
/* It's static to save time */
static unsigned char *tmpgroup=NULL; /* GLOBAL to this module */

/* ---------------------------------------------------------------------- */

/* for the X display */
Display       *x_display;
int           x_screen;
Window        x_window;
int           x_depth;
unsigned long x_event_mask;

static GC      x_gc_clear=0;

static Pixmap  x_pixmap;
static XImage  *x_image;

/* Colormap variables */
extern int           c_cmap; /* whether there was a colormap allocated */
extern Colormap      x_cmap;
extern unsigned long pixels[256];
extern int           pal1to1; /* flag: whether cmap index is array index */


/* ====================================================================== */

/* initialize_display : initialize graphics mode with min size widthXheight
** Set the graphics mode
** a nonresizable window of at least width and height is made if possible
**
** Also allocate memory for xscale
** GLOBAL to this module, non dev specific: tmpgroup
*/
void
initialize_display(int width, int height)
{
  char                 *displayname=NULL;
  XGCValues            gcvals;
  XSetWindowAttributes win_attr;
  XSizeHints           hints;
  char                 title[256];

  /* Open the connection to the X server */
  if((x_display=XOpenDisplay(displayname))==NULL) {
    logmesg("init_disp: could not open X display\n");
    bye_bye(1);
  }


  /* Create the window */
  x_screen = DefaultScreen(x_display);
  x_depth = DefaultDepth(x_display, x_screen);
  w_width=width;
  w_height=height;
  if(width<320) w_width=320;
  if(height<200) w_height=200;
  x_window = XCreateSimpleWindow(/* display */ x_display,
               /* parent window */ DefaultRootWindow(x_display),
               /* x , y position */ 100, 100,
               /* width , height */ w_width, w_height,
               /* border_width */ 0,
               /* border color */ BlackPixel(x_display, x_screen),
               /* background color */ BlackPixel(x_display, x_screen)
             );

  /* Set the window attributes */
  /* Receive mouse button and keyboard events */
  win_attr.event_mask = x_event_mask = KeyPressMask | ButtonPressMask;

  /* use backing store because it's easier than dealing with expose events */
  win_attr.backing_store = Always;

  XChangeWindowAttributes(x_display, x_window,
    CWEventMask | CWBackingStore, &win_attr);

  /* Set the window properties */
  sprintf(title, "ShowTime");
  hints.flags = PSize | PMinSize | PMaxSize;
  hints.width = hints.min_width = hints.max_width = w_width;
  hints.height = hints.min_height = hints.max_height = w_height;
  XSetStandardProperties(x_display, x_window, title, title, None,
                         NULL, 0, &hints);

  /* Display the window */
  XMapWindow(x_display, x_window);
  XSync(x_display, /* discard events */ False);

#define FACTOR_OF_4
#ifdef FACTOR_OF_4
  /* Force the ximage width to be a factor of 4, because wraparound
  ** happens in some versions of X otherwise.
  */
  width=((int)(w_width/4.0 + 0.9999))*4;
#endif
  /* Initialize the one pixmap I'll use for all the images */
  x_pixmap = XCreatePixmap(x_display, x_window, width, w_height, x_depth);

  /* Create a "clear screen" graphics context */
  gcvals.foreground = BlackPixel(x_display, x_screen);
  x_gc_clear = XCreateGC(x_display, x_window, GCForeground, &gcvals);

  /* Create a blank XImage to draw in */
  XFillRectangle(x_display, x_pixmap, x_gc_clear, 0, 0, w_width, w_height);

  x_image = XGetImage(x_display, /* drawable */ x_pixmap,
               /* x, y offset */ 0, 0,
               /* width, height */ width, w_height,
               /* plane mask */ AllPlanes,
               /* format */ ZPixmap);

  /* tmpgroup is GLOBAL to module */
  /* Allocate some memory for the image manipulation buffer */
  if((tmpgroup=malloc((size_t)128 + (size_t)w_width))==NULL) {
    logmesg("initialize_display: could not allocate xscale buffer\n");
  }
}


























/* ---------------------------------------------------------------------- */
/* Clear the window */
void
clear_display(int bgcolor)
{
  XGCValues            gcvals;

  /* Create a "clear screen" graphics context */
  /*gcvals.foreground = BlackPixel(x_display, x_screen);/**/
  gcvals.foreground = pixels[bgcolor];

  /*x_gc_clear = XCreateGC(x_display, x_window, GCForeground, &gcvals);/**/
  XChangeGC(x_display, x_gc_clear, GCForeground, &gcvals);

  XFillRectangle(x_display, x_pixmap, x_gc_clear, 0, 0, w_width, w_height);
  XCopyArea(x_display, /*src*/x_pixmap, /*dest*/x_window,
            DefaultGC(x_display, x_screen),
            /* src */ 0, 0, /* size */ w_width, w_height, /* dest */ 0, 0);
  XSync(x_display, /* discard events */False);
}

/* ---------------------------------------------------------------------- */

/* reset_display : put the graphics mode into a normal text mode
*/
void
reset_display(void)
{
  if(tmpgroup!=NULL) free(tmpgroup);
  if(c_cmap) XFreeColormap(x_display, x_cmap);
  if(x_gc_clear) XFreeGC(x_display, x_gc_clear);
}




/* ====================================================================== */

static int buffered_image=0;

/* show_image : scale and display an image loaded into memory
*/
void
show_image(screendata_t *sdP)
{
  register int           hi;
  register int           xsize; /* screen (destination) image width */
  register unsigned char *newgroup;
  register char          *xdata=x_image->data;
  register int           xi_width=x_image->width;
  register char          *dest_row;
  register int           ysize; /* screen (destination) image height */
  register unsigned char *sd_row;
  register int           prev_src_row;
  register int           src_row;
  register int           xstep;
  register int           ystep;
  register int           vi;

  if(sdP==NULL) return;

  if(gm_xscale!=1) {
    newgroup=tmpgroup;
  }

  if(gm_xscale>0) {
    xsize = sdP->num_cols * gm_xscale;
  } else if(gm_xscale<0) {
    xstep = -gm_xscale;
    xsize = sdP->num_cols / xstep;
  } else {
    xsize = sdP->num_cols;
  }

  if(gm_yscale>0) {
    ysize = sdP->num_rows * gm_yscale;
  } else if(gm_yscale<0) {
    ystep = -gm_yscale;
    ysize = sdP->num_rows / ystep;
  } else {
    ysize = sdP->num_rows;
  }

  /* Clip the image if it overwrites the window */
  if(xsize + gm_xoff > w_width) xsize = w_width - gm_xoff;
  if(ysize + gm_yoff > w_height) ysize = w_height - gm_yoff;
  if(xsize<1 || ysize<1) return;

  /* Display each row of the image */
  prev_src_row = -1;
  for(vi=0; vi<ysize; vi++) {

    if(gm_yscale > 0) {
      src_row = (int)(vi/gm_yscale);
    } else {
      src_row = vi*ystep;
    }
    sd_row = sdP->data[src_row];

    if(src_row != prev_src_row) {
      /* New source image row */
      /* Scale horizontally */
      if(gm_xscale!=1 && gm_xscale>0) {
        for(hi=0; hi<xsize; hi++) {
          newgroup[hi] = sd_row[sdP->header_size + (int)(hi/gm_xscale)];
        }
      } else if(gm_xscale!=1 && gm_xscale<0) {
        /* shrink integrally */
        for(hi=0; hi<xsize; hi++) {
          newgroup[hi] = sd_row[sdP->header_size + hi*xstep];
        }
      } else {
        newgroup=sd_row;
      }

      /* Put the image up on the graphics screen */
      /* This part is the most CPU intensive of the whole code */
      if(x_depth == 8) {
        if(pal1to1) {
          memcpy(/*to*/xdata+vi*xi_width, /*from*/newgroup, xsize);
        } else {
          dest_row=xdata+vi*xi_width;
          for(hi=0; hi<xsize; hi++) {
            *(dest_row+hi) = pixels[*(newgroup+hi)];
          }
        }
      } else {
        logmesg("depth %i not supported\n", x_depth);
      }
    } else {
      /* Current source row is the same as the previous */
      /* Simply duplicate the previous source image row */
      memcpy(/*to*/xdata+vi*xi_width, /*from*/xdata+(vi-1)*xi_width, xsize);
    }

    prev_src_row=src_row;
  } /* for each row of the screen (destination) image */

  /* Copy the new XImage into the pixmap */
  /* This command sends the image to the X server */
  XPutImage(x_display, /*drawable*/x_pixmap, DefaultGC(x_display, x_screen),
            x_image, /*src*/ 0, 0, /*dest*/ gm_xoff, gm_yoff,
            /*size*/ xsize, ysize);

  /* Only update the window if we're not inside an xform */
  /* This reduces the strobing effect which somes from the small delay
  ** between XPutImage and XCopyArea.
  */
  if(!gm_in_do_xform) {
    if(buffered_image) {
      /* Copy whole pixmap to window */
      /* This command draws the pixmap on the window */
      XCopyArea(x_display, /*src*/x_pixmap, /*dest*/x_window,
                DefaultGC(x_display, x_screen),
                /* src */ 0, 0,
                /* size */ w_width, w_height,
                /* dest */ 0, 0);
      buffered_image = 0;
    } else {
      /* NOTE: To to a better job of handling expose events, especially
      **   in the situation where Backing Store is not available,
      **   change the XCopyArea command to copy the whole pixmap to the
      **   window, by setting src and dest to 0,0 and size to w_width,w_height
      */
      /* Copy only new portion of pixmap to window */
      /* This command draws the pixmap on the window */
      XCopyArea(x_display, /*src*/x_pixmap, /*dest*/x_window,
                DefaultGC(x_display, x_screen),
                /* src */ gm_xoff, gm_yoff,
                /* size */ xsize, ysize,
                /* dest */ gm_xoff, gm_yoff);
    }
  } else {
    /* If we're in an xform, notify ourselves of the updated pixmap
    ** in the next call to show_image
    */
    buffered_image = 1;
  }

  /* Make sure the window gets updated right now */
  XSync(x_display, /* discard events */ False);

  return;
}
